import sys
import requests
from bs4 import BeautifulSoup
from urllib.parse import urljoin, urlparse
import re
import json
import time
from requests.exceptions import Timeout

def scrape(url, timeout=10):
    # Start timing the scraping process
    start_time = time.time()

    try:
        # Ensure the URL is complete
        parsed_url = urlparse(url)
        if not parsed_url.scheme:
            url = 'https://' + url
            
        base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"

        # Add timeout to the request
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()

        soup = BeautifulSoup(response.content, 'html.parser')

        # Check if we're still within timeout
        if time.time() - start_time > timeout:
            return json.dumps({
                'success': False,
                'message': f"Scraping took too long (>{timeout}s)",
                'InputURL': url
            })

        text_content = soup.get_text(separator=' ', strip=True)
        html_content = str(soup)
        
        # Store all results from all pages
        all_emails = set()
        all_social_media_links = set()
        all_phone_numbers = set()
        visited_urls = {url.lower()}
        
        # Process main page first
        main_page_data = process_page_content(soup, html_content, text_content, url, timeout, start_time)
        if not main_page_data['success']:
            return json.dumps(main_page_data)
            
        all_emails.update(main_page_data['emails'])
        all_social_media_links.update(main_page_data['social_media_links'])
        all_phone_numbers.update(main_page_data['phone_numbers'])
        
        # Extract links from the main page
        priority_pages = find_priority_pages(soup, base_url, url)
        
        # Visit up to 3 priority pages
        for i, page_url in enumerate(priority_pages[:3]):
            # Check if we still have time
            if time.time() - start_time > timeout * 0.7:  # Use 70% of timeout for main + subpages
                break
                
            try:
                # Skip if we've already visited this URL
                if page_url.lower() in visited_urls:
                    continue
                    
                visited_urls.add(page_url.lower())
                
                # Add a small delay to avoid rate limiting
                time.sleep(0.5)
                
                # Get the page content
                sub_response = requests.get(page_url, timeout=timeout/4)  # Shorter timeout for subpages
                sub_response.raise_for_status()
                
                sub_soup = BeautifulSoup(sub_response.content, 'html.parser')
                sub_text_content = sub_soup.get_text(separator=' ', strip=True)
                sub_html_content = str(sub_soup)
                
                # Process the subpage
                subpage_data = process_page_content(sub_soup, sub_html_content, sub_text_content, page_url, timeout, start_time)
                if subpage_data['success']:
                    all_emails.update(subpage_data['emails'])
                    all_social_media_links.update(subpage_data['social_media_links'])
                    all_phone_numbers.update(subpage_data['phone_numbers'])
            
            except Exception as e:
                # Just continue if a subpage fails
                continue
        
        # Get image list from the main page only
        image_list = extract_images(soup, url, timeout, start_time)
        
        # Calculate final timing
        end_time = time.time()
        scraping_time = end_time - start_time
        
        # Convert sets to sorted lists and keep the quality scoring for emails
        # Use the email_quality_scores dictionary from the main_page_data
        emails_with_scores = [(email, main_page_data['email_quality_scores'].get(email, 0)) for email in all_emails]
        # Sort by quality score
        emails_with_scores.sort(key=lambda x: x[1], reverse=True)
        sorted_emails = [email for email, score in emails_with_scores]

        return json.dumps({
            "success": True,
            "text_content": text_content[:100000],  # Limit text content size
            "html_content": html_content[:200000],  # Limit HTML content size
            "emails": sorted_emails[:20],  # Limit number of emails
            "phone_numbers": list(all_phone_numbers)[:20],  # Limit number of phone numbers
            "social_media_links": list(all_social_media_links)[:20],  # Limit number of social links
            "image_list": image_list[:50],  # Limit number of images
            "visited_pages": list(visited_urls),  # List of pages visited
            "scraping_time": scraping_time,
            "InputURL": url
        })

    except Timeout:
        return json.dumps({
            'success': False,
            'message': f"Request timed out after {timeout} seconds",
            'InputURL': url
        })
    except requests.exceptions.RequestException as e:
        return json.dumps({
            'success': False,
            'message': f"Error fetching URL: {str(e)}",
            'InputURL': url
        })
    except Exception as e:
        return json.dumps({
            'success': False,
            'message': f"Unexpected error: {str(e)}",
            'InputURL': url
        })

def find_priority_pages(soup, base_url, current_url):
    """
    Find and prioritize pages that are likely to contain contact information.
    Returns a list of URLs ordered by likelihood of containing contact info.
    """
    links = soup.find_all('a', href=True)
    pages = []
    
    # Keywords for each priority level
    contact_keywords = ['contact', 'kontakt', 'contacto', 'contato', 'contacte', 'get-in-touch', 'get_in_touch', 'reach-us', 'connect']
    about_keywords = ['about', 'about-us', 'about_us', 'who-we-are', 'company', 'team', 'staff', 'our-team', 'meet-the-team']
    other_keywords = ['home', 'main', 'index', 'services', 'products', 'portfolio']
    
    for link in links:
        href = link.get('href')
        
        # Skip empty links, javascript links, anchors, etc.
        if not href or href.startswith('#') or href.startswith('javascript:') or href.startswith('mailto:') or href.startswith('tel:'):
            continue
            
        # Get the absolute URL
        abs_url = urljoin(current_url, href)
        
        # Only consider links to the same domain
        if not abs_url.startswith(base_url):
            continue
            
        # Skip links to images, PDFs, etc.
        if abs_url.lower().endswith(('.jpg', '.jpeg', '.png', '.gif', '.pdf', '.doc', '.docx', '.xls', '.xlsx', '.zip', '.tar.gz')):
            continue
            
        # Skip if the link is the current page
        if abs_url == current_url:
            continue
            
        # Score the link based on keywords in the URL and link text
        url_path = urlparse(abs_url).path.lower()
        link_text = link.get_text().lower()
        
        # Skip if already in pages list
        if abs_url in pages:
            continue
            
        # Check for footer links (often contain contact info)
        in_footer = False
        parent = link.parent
        for _ in range(5):  # Check up to 5 levels up
            if parent and parent.name:
                if parent.name.lower() in ['footer', 'foot', 'bottomnav'] or parent.get('id', '').lower() in ['footer', 'foot', 'bottom'] or 'footer' in parent.get('class', []):
                    in_footer = True
                    break
            if parent:
                parent = parent.parent
            else:
                break
                
        # Assign a priority score (lower is better)
        priority = 100
        
        # URL contains contact keywords
        if any(keyword in url_path for keyword in contact_keywords):
            priority = 1
        # Link text contains contact keywords
        elif any(keyword in link_text for keyword in contact_keywords):
            priority = 2
        # In footer and contains contact/about keywords
        elif in_footer and (any(keyword in url_path for keyword in contact_keywords + about_keywords) or 
                           any(keyword in link_text for keyword in contact_keywords + about_keywords)):
            priority = 5
        # URL contains about keywords
        elif any(keyword in url_path for keyword in about_keywords):
            priority = 10
        # Link text contains about keywords
        elif any(keyword in link_text for keyword in about_keywords):
            priority = 15
        # In footer
        elif in_footer:
            priority = 20
        # Other interesting pages
        elif any(keyword in url_path for keyword in other_keywords):
            priority = 30
        
        pages.append((priority, abs_url))
    
    # Sort by priority and return just the URLs
    pages.sort()
    return [url for _, url in pages]

def extract_images(soup, url, timeout, start_time):
    """
    Extract images from a page.
    """
    images = soup.find_all('img')
    image_list = []
    for img in images:
        if time.time() - start_time > timeout:
            return image_list
        
        src = img.get('src')
        if src:  # Only process if src exists
            src = urljoin(url, src)
            image_list.append(src)
    
    return image_list

def process_page_content(soup, html_content, text_content, url, timeout, start_time):
    """
    Process the content of a page to extract emails, phone numbers, and social media links.
    Returns a dictionary with the extracted information.
    """
    # Enhanced email patterns
    email_patterns = [
        # Standard email pattern
        r'[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}',
        # Pattern with possible surrounding characters
        r'[\w\.-]+@[\w\.-]+\.\w+',
        # Pattern that might be in mailto: links
        r'mailto:([a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,})',
        # Pattern that might be encoded in URLs
        r'(?:mailto:|%40)([a-zA-Z0-9._%+-]+(?:@|%40)[a-zA-Z0-9.-]+\.[a-zA-Z]{2,})',
        # Pattern for emails with special characters
        r'[a-zA-Z0-9][a-zA-Z0-9._%+-]*@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}',
        # Pattern for emails in text content
        r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'
    ]

    # Exclusion patterns for false positives
    exclude_patterns = [
        r'@sentry\.wixpress\.com$',
        r'@sentry-next\.wixpress\.com$',
        r'@sentry\.io$',
        r'^[a-f0-9]{32}@',
        r'^user@domain\.com$',
        r'contenthash\.[a-f0-9]+\.(webp|jpg|jpeg|png|gif|svg)$',
        r'^info@company\.com$',
        r'^address@email\.com$',
        r'^your@domain\.com$',
        r'^example@mysite\.com$',
        r'^u003e',
        r'@2x\.',
        r'@3x\.',
        r'@4x\.',
        r'^email@example\.com$',
        r'^email@domain\.com$',
        r'^team@latofonts\.com$',
        r'^filler@godaddy\.com$',
        r'@.*\.(png|jpg|jpeg|gif|svg)$',
        r'typed\.js@',
        r'-.*@.*\.(png|jpg|jpeg|gif|svg)$',
        r'^someone@example\.com$',
        r'^.*@example\.com$',
        # New exclusions
        r'tailwindcss@[\d\.]+',
        r'slick-carousel@[\d\.]+',
        r'^johndoe@email\.com$',
        r'isotope-layout@[\d\.]+',
        r'^info@yourwebsite\.com$',
        r'^support@rstheme\.com$',
        r'fancybox@[\d\.]+',
        r'canvg@[\d\.]+',
        r'core-js-bundle@[\d\.]+',
        r'focus-within-polyfill@[\d\.]+',
        r'lodash@[\d\.]+',
        r'custom-select@[\d\.]+',
        r'holy-stars@[\d\.]+',
        # Additional exclusions for common false positives and package references
        r'@[\d\.]+',  # Version numbers
        r'@import',  # CSS imports
        r'@media',  # CSS media queries
        r'@font-face',  # CSS font-face
        r'@keyframes',  # CSS animations
        r'^noreply@',  # Generic noreply emails
        r'^no-reply@',  # Generic no-reply emails
        r'^donotreply@',  # Generic do not reply emails
        r'^webmaster@',  # Generic webmaster emails
        r'^admin@',  # Generic admin emails
        r'^test@',  # Test emails
        r'^info@example',  # Example info emails
        r'^contact@example',  # Example contact emails
        r'^sales@example',  # Example sales emails
        r'^support@example',  # Example support emails
        r'^[a-z0-9]{1,3}@',  # Very short usernames (likely false positives)
        r'@localhost',  # Local development
        r'@127.0.0.1',  # Local development IP
        r'@test\.',  # Test domains
        r'@example\.',  # Example domains
        r'\.example\.',  # Example subdomains
        r'@demo\.',  # Demo domains
        r'@sample\.',  # Sample domains
        r'@.*\.(test|example|invalid|localhost|local)$',  # Invalid TLDs
        r'jquery@[\d\.]+',  # jQuery versions
        r'bootstrap@[\d\.]+',  # Bootstrap versions
        r'react@[\d\.]+',  # React versions
        r'vue@[\d\.]+',  # Vue versions
        r'angular@[\d\.]+',  # Angular versions
        r'npm@[\d\.]+',  # npm versions
        r'webpack@[\d\.]+',  # Webpack versions
        r'gatsby@[\d\.]+',  # Gatsby versions
        r'next@[\d\.]+',  # Next.js versions
        r'swiper@[\d\.]+',  # Swiper versions
        r'rollup@[\d\.]+',  # Rollup versions
        r'@fortawesome',  # Font Awesome
        r'@babel',  # Babel
        r'@testing-library',  # Testing library
        r'@microsoft',  # Microsoft packages
        r'@types',  # TypeScript types
        r'@aws',  # AWS packages
        r'@vercel',  # Vercel packages
        r'@[\d\w]+\.[\d\w]+\.cdn',  # CDN references
        r'@(google|facebook|twitter|apple|amazon)\.com',  # Common fake domains
    ]
    
    # Common disposable email domains for filtering
    disposable_domains = [
        'mailinator.com', 'yopmail.com', 'guerrillamail.com', 'sharklasers.com',
        'mailnesia.com', 'temporary-mail.net', '10minutemail.com', 'tempmail.com',
        'throwawaymail.com', 'spamgourmet.com', 'getairmail.com', 'fakeinbox.com',
        'dispostable.com', 'getnada.com', 'emailondeck.com', 'trashmail.com',
        'maildrop.cc', 'tempr.email', 'tempmail.org', 'emailfake.com',
        'fakemailgenerator.com', 'tempmailo.com', 'temp-mail.org',
    ]

    # Common top-level domains for quality scoring
    high_quality_tlds = [
        'com', 'net', 'org', 'edu', 'gov', 'io', 'co', 'ai', 'app', 'dev',
        'me', 'info', 'us', 'uk', 'ca', 'au', 'de', 'fr', 'jp', 'it', 'nl',
        'es', 'ch', 'se', 'no', 'dk', 'fi', 'nz', 'sg', 'hk', 'ie', 'br'
    ]

    email_set = set()
    email_quality_scores = {}
    
    # Check if we're still within timeout
    if time.time() - start_time > timeout:
        return {
            'success': False,
            'message': f"Email extraction exceeded timeout ({timeout}s)",
            'InputURL': url
        }
    
    # Search in both HTML and text content
    for content in [html_content, text_content]:
        for pattern in email_patterns:
            found_emails = re.findall(pattern, content, re.IGNORECASE)
            for email in found_emails:
                # Clean up the email if it's from a mailto: or encoded
                if isinstance(email, tuple):
                    email = email[0]
                email = email.lower().strip()
                email = email.replace('%40', '@')
                if email.startswith('mailto:'):
                    email = email[7:]
                
                # Basic email structure validation
                if not (
                    '@' in email and  # Must contain @
                    '.' in email.split('@')[1] and  # Domain must contain .
                    len(email) >= 5 and  # Minimum length
                    not email.startswith('@') and  # Can't start with @
                    not email.endswith('@') and  # Can't end with @
                    not '..' in email and  # No consecutive dots
                    not '@.' in email and  # No dot right after @
                    not '.@' in email  # No dot right before @
                ):
                    continue
                    
                # Check against exclusion patterns
                if any(re.search(exclude, email, re.IGNORECASE) for exclude in exclude_patterns):
                    continue
                    
                # Parse the email
                try:
                    username, domain = email.split('@', 1)
                except ValueError:
                    continue
                    
                # More advanced username validation
                if (
                    len(username) < 2 or  # Username too short
                    username.isdigit() or  # Username is only numbers
                    re.match(r'^[0-9a-f]{6,}$', username) or  # Likely a hash
                    re.match(r'^[\d\.-_]+$', username) or  # Only digits and symbols
                    username.count('.') > 3 or  # Too many dots
                    username.count('_') > 3 or  # Too many underscores
                    username.count('-') > 3 or  # Too many hyphens
                    re.match(r'^\d+\w*$', username)  # Starts with numbers only
                ):
                    continue
                    
                # Advanced domain validation
                try:
                    domain_parts = domain.split('.')
                    tld = domain_parts[-1]
                    
                    # Invalid TLD checks
                    if (
                        len(tld) < 2 or  # TLD too short
                        tld.isdigit() or  # TLD is only numbers
                        len(domain_parts) < 2 or  # Domain must have at least 2 parts
                        domain in disposable_domains or  # Disposable email domain
                        re.match(r'.*\.(png|jpg|jpeg|gif|svg|webp|css|js)$', domain, re.IGNORECASE)  # File extensions
                    ):
                        continue
                        
                    # Compute a quality score for the email (higher is better)
                    quality_score = 0
                    
                    # Username quality factors
                    quality_score += min(len(username), 12) * 0.5  # Longer usernames (up to a point)
                    quality_score -= username.count('.') * 1  # Penalize dots
                    quality_score -= username.count('_') * 1  # Penalize underscores
                    quality_score -= username.count('-') * 1  # Penalize hyphens
                    quality_score -= sum(c.isdigit() for c in username) * 0.5  # Penalize digits
                    
                    # Domain quality factors
                    quality_score += 10 if tld.lower() in high_quality_tlds else 0  # Bonus for common TLDs
                    quality_score += 5 if len(domain_parts) == 2 else 0  # Simple domains often better
                    quality_score -= 5 if len(domain_parts) > 3 else 0  # Penalize complex domains
                    quality_score -= 10 if 'temp' in domain or 'fake' in domain else 0  # Penalize suspicious domains
                    
                    # Context quality factors - extra points for business context
                    quality_score += 15 if email in html_content and email in text_content else 0  # Bonus for appearing in both
                    quality_score += 10 if f"mailto:{email}" in html_content else 0  # Bonus for mailto links
                    quality_score += 5 if re.search(r'contact|email|reach|mail', content.lower()) and email in content else 0  # Context clues
                    
                    # Business emails are more valuable
                    business_keywords = ['contact', 'info', 'sales', 'support', 'hello', 'admin', 'marketing', 
                                         'team', 'help', 'enquiry', 'inquiry', 'office', 'customer']
                    if any(keyword in username for keyword in business_keywords):
                        quality_score += 15  # Big bonus for business email patterns
                    
                    # Bonus for emails on contact/about pages
                    if any(keyword in url.lower() for keyword in ['contact', 'about', 'team', 'staff']):
                        quality_score += 10  # Bonus for emails on contact/about pages
                    
                    email_set.add(email)
                    email_quality_scores[email] = quality_score
                    
                except Exception:
                    continue

    # Check if we're still within timeout
    if time.time() - start_time > timeout:
        return {
            'success': False,
            'message': f"Phone number extraction exceeded timeout ({timeout}s)",
            'InputURL': url
        }
        
    # Extract phone numbers
    phone_patterns = [
        r'\b\d{3}[-.]?\d{3}[-.]?\d{4}\b',  # Standard US format
        r'\b\(\d{3}\)\s*\d{3}[-.]?\d{4}\b',  # (123) 456-7890
        r'\b\+\d{1,3}[-. ]?\d{3}[-. ]?\d{3}[-. ]?\d{4}\b'  # International
    ]
    phone_set = set()
    for pattern in phone_patterns:
        phone_set.update(re.findall(pattern, html_content))
    phone_numbers = list(phone_set)

    # Check if we're still within timeout
    if time.time() - start_time > timeout:
        return {
            'success': False,
            'message': f"Social media extraction exceeded timeout ({timeout}s)",
            'InputURL': url
        }
        
    # Extract social media links - expanded to include more platforms
    social_media_patterns = [
        "facebook.com", "twitter.com", "instagram.com", "linkedin.com",
        "youtube.com", "tiktok.com", "pinterest.com", "snapchat.com",
        "discord.gg", "t.me", "telegram.me", "wa.me", "whatsapp.com",
        "github.com", "medium.com", "reddit.com", "tumblr.com",
        "vimeo.com", "flickr.com", "behance.net", "dribbble.com", 
        "twitch.tv", "vk.com", "weibo.com", "line.me", "quora.com",
        "meetup.com", "soundcloud.com", "bandcamp.com"
    ]
    social_media_set = set()
    
    # First check <link> tags which often have canonical social media links
    for link in soup.find_all('link', rel=True, href=True):
        href = link['href']
        if any(social_media in href.lower() for social_media in social_media_patterns):
            social_media_set.add(href)
    
    # Then check regular <a> links
    for link in soup.find_all('a', href=True):
        href = link['href']
        # Get absolute URL
        abs_href = urljoin(url, href)
        if any(social_media in abs_href.lower() for social_media in social_media_patterns):
            social_media_set.add(abs_href)
            
    # Also check for social media links in meta tags (OpenGraph, Twitter cards)
    for meta in soup.find_all('meta', content=True):
        content = meta.get('content', '')
        if any(social_media in content.lower() for social_media in social_media_patterns):
            if re.match(r'^https?://', content):
                social_media_set.add(content)
    
    social_media_links = list(social_media_set)

    # Sort emails by quality score
    emails = sorted(list(email_set), key=lambda e: email_quality_scores.get(e, 0), reverse=True)
    
    return {
        'success': True,
        'emails': emails,
        'phone_numbers': phone_numbers,
        'social_media_links': social_media_links,
        'email_quality_scores': email_quality_scores,
        'text_content': text_content,
        'html_content': html_content
    }

if __name__ == "__main__":
    if len(sys.argv) > 1:
        url = sys.argv[1]
        timeout = float(sys.argv[2]) if len(sys.argv) > 2 else 10
        print(scrape(url, timeout))
    else:
        print(json.dumps({'success': False, 'message': 'URL is required'}))